########################################################################
# metaf2xml/XML.pm 1.42
#   write data of a METAR, TAF, or SYNOP message as XML
#
# copyright (c) metaf2xml 2006-2010
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
########################################################################

package metaf2xml::XML;

########################################################################
# some things strictly perl
########################################################################
use strict;
use warnings;

########################################################################
# export the functions provided by this module
########################################################################
BEGIN {
    use Exporter();
    our ($VERSION, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);

    $VERSION     = '1.42';
    @ISA         = qw(Exporter);
    @EXPORT      = ();
    %EXPORT_TAGS = ();
    @EXPORT_OK   = qw(start_xml print_report finish_xml);
}
our @EXPORT_OK;

END { }

=head1 NAME

metaf2xml::XML

=head1 SYNOPSIS

  use metaf2xml::XML;
  metaf2xml::XML::start_xml(\%opts);
  metaf2xml::XML::print_report(\%report);
  ...
  metaf2xml::XML::finish_xml();

=head1 DESCRIPTION

This module writes the data obtained from parsing METAR, TAF, and/or SYNOP
messages as XML.

=head1 SUBROUTINES/METHODS

=cut

########################################################################
# global variables
########################################################################
# number of spaces to prepend. -1: file not open
my $indent_depth = -1;
my $had_first_report;

########################################################################
# helper functions to write XML
########################################################################

sub _XMLtag {
    my $type = shift;
    my $tag = shift;

    print XML ' ' x $indent_depth . '<' . $tag;
    for (my $i = 0; $i < $#_; $i += 2) {
        my ($name, $value) = @_[$i, $i + 1];
        $value =~ s/&/&amp;/g; # must be the first substitution
        $value =~ s/</&lt;/g;
        $value =~ s/>/&gt;/g;
        $value =~ s/"/&quot;/g;
        print XML ' ' . $name . '="' . $value . '"';
    }
    if ($type eq 'start') {
        $indent_depth++;
    } else {
        print XML '/';
    }
    print XML ">\n";
    return;
}

sub _XMLstartTag {
    return _XMLtag 'start', @_;
}

sub _XMLendTag {
    print XML ' ' x --$indent_depth . '</' . $_[0] . '>' . "\n";
    return;
}

sub _XMLemptyTag {
    return _XMLtag 'empty', @_;
}

########################################################################
# helper functions
########################################################################

sub _printXML;
sub _printXML {
    my ($r, $tag, $xml_tag) = @_;

    return unless exists $r->{$tag};
    $r = $r->{$tag};

    $tag = $xml_tag if $xml_tag;
    if (ref $r eq 'HASH') {
        my (@subnodes, @attrs);

        # sort, so XML looks nicer
        for (qw(occurred errorType warningType s v rp rn u q)) {
            push @attrs, $_, $r->{$_} if exists $r->{$_};
        }

        # special tag sequence to comply with the sequence of tags in DTD
        if ($tag eq 'cloudOpacityLvl') {
            @subnodes = qw(locationAnd oktas weather cloudType cloudBase);
        } elsif ($tag eq 'reportModifier') {
            @subnodes = qw(modifierType isLastSegment segment bulletinSeq sequenceLost over24hLate);
        } elsif ($tag eq 'phenomenon') {
            @subnodes = qw(phenomDescrPre phenomDescrPost weather cloudType cloudCover lightningType otherPhenom obscgMtns locationAnd MOV MOVD isStationary cloudTypeAsoctd cloudTypeEmbd);
        } elsif ($tag eq 'cloudInfo') {
            @subnodes = qw(cloudBaseNotAvailable cloudBase cloudBaseFrom cloudBaseTo visVert visVertFrom visVertTo cloudTypeNotAvailable cloudType cloudOktas);
        } elsif ($tag =~ /^(?:radShortWave|radDirectSolar)$/) {
            @subnodes = qw(radiationPeriod notAvailable radiationValue);
        } elsif ($tag eq 'location') {
            @subnodes = qw(inDistance inVicinity distance isQuadrant locationSpec compassDir quadrant);
        } elsif ($tag eq 'climate') {
            @subnodes = qw(traces precipAmount1Inch precipAmount1MM precipAmount2Inch precipAmount2MM temp1 temp2);
        } elsif ($tag eq 'precipHourly') {
            @subnodes = qw(notAvailable timePeriod timeBeforeObs timeSince time_var_Arr precipAmount traces);
        } elsif ($tag eq 'snowFall' || $tag eq 'waterEquivOfSnow') {
            @subnodes = qw(timePeriod timeBeforeObs timeSince time_var_Arr noMeasurement precipAmount traces);
        } elsif ($tag eq 'highestMeanSpeed') {
            @subnodes = qw(measurePeriod timePeriod timeAt timeBeforeObs time_var_Arr wind);
        } elsif ($tag eq 'seaLandIce') {
            @subnodes = qw(notAvailable text seaIceConcentration iceDevelopmentStage iceOfLandOrigin iceEdgeBearing iceConditionTrend);
        } elsif ($tag eq 'maxWindForce') {
            @subnodes = qw(timeBeforeObs notAvailable windForce);
        } else {
            for (sort keys %$r) {
                if (!/^(?:[vuqs]|rp|rn|errorType|warningType|occurred)$/) {
                    push @subnodes, $_;
                }
            }
        }

        # *Arr are arrays with different subnodes. Suppress tag in XML.
        if ($#subnodes > -1) {
            _XMLstartTag $tag, @attrs unless $tag =~ /Arr$/;
            for (@subnodes) {
                _printXML $r, $_;
            }
            _XMLendTag $tag unless $tag =~ /Arr$/;
        } else {
            _XMLemptyTag $tag, @attrs unless $tag =~ /Arr$/;
        }
    } elsif (ref $r eq 'ARRAY') {
        if ($#$r > -1) {
            for (@$r) {
                _printXML { $tag => $_ }, $tag;
            }
        } else {
            _XMLemptyTag $tag;
        }
    } else {
        _XMLemptyTag $tag, defined $r ? ('v',  $r) : ();
    }
    return;
}

sub _cmpVersions {
    my $version = shift;

    if ($version ne '1.42') {
        _printXML { ERROR => { errorType => 'versionsDiffer',
                               s         =>
                                  "parser.pm: '$version' <@> XML.pm: '" . '1.42'
                             }
                  }, 'ERROR';
        return 0;
    }
    return 1;
}

########################################################################
# start_xml
########################################################################

=head2 start_xml(\%opts)

This function opens the file and writes to it the optional data as specified.

The following argument is expected:

=over

=item C<opts>

Reference to hash of options. The following keys of the hash are recognised:

=over

=item C<x>, with value B<E<lt>xml_fileE<gt>>

enables writing the data to <xml_file>

=item C<D>

include DOCTYPE and reference to the DTD

=item C<S>, with value B<E<lt>xslt_fileE<gt>>

include reference to the stylesheet <xslt_file>

=item C<O>, with value B<E<lt>optionsE<gt>>

include <options> (a space separated list of options)

=back

Without the key C<x>, no output is generated.
The keys C<D>, C<S>, C<O> imply C<X>.

=back

=cut

sub start_xml {
    my $opts = shift;

    return 0
        unless $indent_depth == -1 && exists $opts->{x};

    # open file, set state
    open(XML, ">$opts->{x}") || return 0;
    $indent_depth = 0;
    $had_first_report = 0;

    # XML header
    print XML '<?xml version="1.0" encoding="UTF-8"?>' . "\n";

    # write additional info to XML

    # XML meta data
    print XML '<!DOCTYPE data SYSTEM "metaf.dtd">' . "\n"
        if exists $opts->{D};
    print XML '<?xml-stylesheet href="' . $opts->{S} . '" type="text/xsl"?>'
              . "\n"
        if exists $opts->{S};

    print XML "\n";

    # opening data tag
    _XMLstartTag 'data';

    if (exists $opts->{O}) {
        my (%options, $msg);
        (@options{'type_metaf', 'type_synop', 'lang', 'format', 'src_metaf', 'src_synop', 'mode', 'hours'}, $msg)
                                                 = split / /, $opts->{O}, 9;
        @options{'msg_metaf', 'msg_synop'} = split '  ', $msg, 2;
        _XMLemptyTag 'options', map { $_ => $options{$_} } keys %options;
    }
    return 1;
}

########################################################################
# print_report
########################################################################

=head2 print_report(\%report)

This function writes the content of C<report> to the XML file.
The data for each message will be enclosed in a tag C<<metar>>, C<<taf>>, or
C<<synop>>. The function can be invoked repeatedly.
L<start_xml()|start_xml___opts_> must have been invoked already.
L<finish_xml()|finish_xml__> must be invoked to complete the XML file.

The following argument is expected:

=over

=item C<report>

reference to hash of data to be written to the file

=back

=cut

sub print_report {
    my $report = shift;

    return 0
        unless $indent_depth >= 0;

    $report->{version} = '0' unless exists $report->{version};
    _XMLstartTag 'reports',
            xmlns       => 'http://metaf2xml.sourceforge.net/1.42',
            query_start => gmtime().'',
        unless $had_first_report;
    $had_first_report = 1;

    if (exists $report->{isSynop}) {
        _XMLstartTag 'synop', 's', $report->{msg};

        if (!_cmpVersions $report->{version}) {
            _XMLendTag 'synop';
            return 0;
        }

        for (qw(ERROR warning
                obsStationType callSign obsTime reportModifier
                windIndicator stationPosition obsStationId precipInd wxInd
                baseLowestCloud visPrev visibilityAtLoc totalCloudCover
                sfcWind temperature stationPressure SLP gpSurface
                pressureChange precipHourly weatherSynop cloudTypes
                exactObsTime))
        {
            _printXML $report, $_;
        }

        if (exists $report->{section2}) {
            if ($#{$report->{section2}} > -1) {
                _XMLstartTag 'section2', s => '222';
                _printXML { section2Arr => $report->{section2}}, 'section2Arr';
                _XMLendTag 'section2';
            } else {
                _XMLemptyTag 'section2', s => '222';
            }
        }

        if (exists $report->{section3}) {
            if ($#{$report->{section3}} > -1) {
                _XMLstartTag 'section3', s => '333';
                _printXML { section3Arr => $report->{section3}}, 'section3Arr';
                _XMLendTag 'section3';
            } else {
                _XMLemptyTag 'section3', s => '333';
            }
        }

        if (exists $report->{section4}) {
            if ($#{$report->{section4}} > -1) {
                _XMLstartTag 'section4', s => '444';
                _printXML { section4Arr => $report->{section4}}, 'section4Arr';
                _XMLendTag 'section4';
            } else {
                _XMLemptyTag 'section4', s => '444';
            }
        }

        if (exists $report->{section5}) {
            if ($#{$report->{section5}} > -1) {
                _XMLstartTag 'section5', s => '555';
                _printXML { section5Arr => $report->{section5}}, 'section5Arr';
                _XMLendTag 'section5';
            } else {
                _XMLemptyTag 'section5', s => '555';
            }
        }

        _XMLendTag 'synop';
    } else {
        my $is_taf;

        $is_taf = exists $report->{isTaf};
        _XMLstartTag $is_taf ? 'taf' : 'metar', 's', $report->{msg};

        if (!_cmpVersions $report->{version}) {
            _XMLendTag $is_taf ? 'taf' : 'metar';
            return 0;
        }

        for (qw(ERROR warning isSpeci
                obsStationId obsTime issueTime fcstPeriod reportModifier
                fcstCancelled fcstNotAvbl
                sfcWind windShearLvlArr
                CAVOK visPrev visMin visRwy RVRNO weather cloud visVert
                temperature QNH QFE waterTemp seaCondition
                somePressure cloudMaxCover recentWeather QFF windShear rwyState
                colourCode NEFO_PLAYA RH windAtLoc rwyWind TAFsupplArr))
        {
            _printXML $report, $_;
        }

        if (exists $report->{trend}) {
            for my $td (@{$report->{trend}}) {
                _XMLstartTag 'trend', 's', $td->{s};
                for (qw(trendType timeAt timeFrom timeTill probability
                        sfcWind CAVOK visPrev visVert weather cloud
                        rwyState colourCode TAFsupplArr))
                {
                    _printXML $td, $_;
                }
                _XMLendTag 'trend';
            }
        }

        _printXML $report, 'TAFinfoArr';
        _printXML $report, 'remark', $is_taf ? 'tafRemark' :'remark';

        _XMLendTag $is_taf ? 'taf' : 'metar';
    }
    return 1;
}

########################################################################
# finish_xml
########################################################################

=head2 finish_xml()

This function must be invoked to complete the writing to the file.
L<start_xml()|start_xml___opts_> must have been invoked already.

No arguments are expected.

=cut

sub finish_xml {
    return 0
        unless $indent_depth >= 0;

    _XMLstartTag 'reports',
            xmlns       => 'http://metaf2xml.sourceforge.net/1.42',
            query_start => gmtime().''
        unless $had_first_report;

    _XMLendTag 'reports';

    # closing data tag
    _XMLendTag 'data';

    # set state, close file
    $indent_depth = -1;
    return close XML;
}

=head1 SEE ALSO

=begin html

<a href=parser.pm.html>metaf2xml::parser(3pm)</a>,
<a href=metaf2xml.pl.html>metaf2xml(1)</a>, <!--

=end html

B<metaf2xml::parser>(3pm),
B<metaf2xml>(1),

=for html -->

L<http://metaf2xml.sourceforge.net/>

=head1 COPYRIGHT and LICENSE

copyright (c) 2006-2010 metaf2xml @ L<http://metaf2xml.sourceforge.net/>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA

=cut

1;
